/*
** 快速排序算法的C++实现
** 假设：对N个数字进行升序排序
** swap()函数：交换序列中两个数的位置
** partition()函数：切分待排序序列，使切分后的两个子序列满足左子序列小于右子序列
** Qsort()函数：递归的进行切分操作
*/

void swap(vector<int>& seq, int low, int high) {
    int temp = seq[low];
    seq[low] = seq[high];
    seq[high] =  temp;
}

int partition(vector<int>& seq, int low, int high) {
    int pivotkey = seq[low], pindex = low; // pivotkey为初始值，pindex为初始值的索引
    while (low < high) {
        while (low < high && seq[high] >= pivotkey) high--; // 从右向左扫描，找到第一个小于初始值的数据项
        while (low < high && seq[low] <= pivotkey) low++; // 从左向右扫描，找到第一个大于初始值的数据项
        swap(seq, low, high);   // 交换两个数据项
    }
    swap(seq, pindex, low); // 将初始值存放到正确的位置
    return low;
}

void QSort(vector<int>& seq, int low, int high) {
    if (low < high) {
        int pivot = partition(seq, low, high);
        QSort(seq, low, pivot - 1);  // 对左子序列递归的进行切分操作
        QSort(seq, pivot + 1, high);  // 对右子序列递归的进行切分操作
    }
}

void QuickSort(vector<int>& seq) {
    QSort(seq, 0, seq.size() - 1);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/*
** 三平均分区法
** 只需修改切分函数
*/

#define MAX(a, b)   (((a) > (b) ) ? (a) : (b))
#define MIN(a, b)   (((a) < (b) ) ? (a) : (b))
#define MID(a,b,c)  (MAX(a,b)>c?MAX(MIN(a,b),c):MIN(MAX(a,b),c))

int partition(vector<int>& seq, int low, int high) {
    int median = MID(low,high,(low+high)/2);  // 选择序列最左边、最右边、中间这三个数的中位数
    int s = ( median == seq[low]?low:(median == seq[high]?high:(low+high)/2) ); // 判断中位数的下标
    int pivotkey = seq[s], pindex = low;
    swap(seq,low,s); // 将中位数与最左边的数互换位置
    while (low < high) {
        while (low < high && seq[high] >= pivotkey) high--;
        while (low < high && seq[low] <= pivotkey) low++;
        swap(seq, low, high);
    }
    swap(seq, pindex, low);
    return low;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/*
** 根据区间动态调整排序算法
** 当区间小于一个阈值时停止快速排序，
** 使序列基本有序，然后对整个基本有序的序列进行插入排序
*/

int partition(vector<int>& seq, int low, int high) {
    if(high - low < 15) // 当区间小于一个阈值时停止快速排序
        return -1;
    int pivotkey = seq[low], pindex = low;
    while (low < high) {
        while (low < high && seq[high] >= pivotkey) high--;
        while (low < high && seq[low] <= pivotkey) low++;
        swap(seq, low, high);
    }
    swap(seq, pindex, low);
    return low;
}

void QSort(vector<int>& seq, int low, int high) {
    if (low < high) {
        int pivot = partition(seq, low, high);
        if (pivot + 1) {
            QSort(seq, low, pivot - 1);
            QSort(seq, pivot + 1, high);
        }
    }
}

void QuickSort(vector<int>& seq) {
    QSort(seq, 0, seq.size() - 1);
    insertSort(seq); // 调用快速排序算法对整个基本有序的序列进行排序
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/*
** 维护三个指针：
** 维护一个指针lt使 seq[low,lt-1] 中的元素都小于pivotkey
** 维护一个指针gt使 seq[gt+1,high]中的元素都大于pivotkey
** 维护一个指针i使  seq[lt,i-1]   中的元素都等于pivotkey
** 而seq[i,gt]中的元素都还为确定
** 注意：此算法只在序列中含有大量重复数据时能发挥作用，否则慢于传统的快速排序
*/

void QSort(vector<int>& seq, int low, int high) {
    if (low < high) {
        int lt = low, gt = high, i = low + 1, pivotkey = seq[low];
        while (i <= gt) {
            if(seq[i] < pivotkey)
                swap(seq, lt++, i++);
            else if(seq[i] > pivotkey) //
                swap(seq, i, gt--);
            else i++;
        }
        QSort(seq, low, lt - 1);  // 对左子序列递归的进行切分操作
        QSort(seq, gt + 1, high);  // 对右子序列递归的进行切分操作
    }
}